﻿using System;
using System.Linq;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.DotProlog.Parsing;
using OpcLabs.DotProlog.Terms;

namespace DotPrologTest.Parsing
{
    // ISO/IEC JTC1 SC22 WG17; Conformity Assessment I: Syntax
    // http://www.complang.tuwien.ac.at/ulrich/iso-prolog/conformity_assessment

    [TestClass]
    // ReSharper disable once InconsistentNaming
    public class ReadTermParserTest_Ulrich : ReadTermParserTestBase
    {
        // ReSharper disable InconsistentNaming

        const string Input_1 = @"writeq('\n').";
        const string Input_2 = @"'";
        const string Input_3 = @")";
        const string Input_4 = @".";
        const string Input_5 = @"writeq('" + "\x9" + @"'). % horiz. tab";
        [NotNull] static string Input_6 { get { return @"writeq('" + Environment.NewLine + @"')."; }}

        // ReSharper restore InconsistentNaming


        #region Test Methods
        // ReSharper disable InconsistentNaming

        [TestMethod]
        public void Test_1()
        {
            TestSuccess(Input_1, new NTerm("writeq") {"\x0A"} );
        }

        [TestMethod]
        public void Test_2()
        {
            TestError(Input_2, TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_3()
        {
            TestWait(Input_3);
        }

        [TestMethod]
        public void Test_4()
        {
            TestError(Input_4, TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_5()
        {
            TestError(Input_5, TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_6()
        {
            TestError(Input_6, TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_7()
        {
            TestSuccess(
                @"writeq('\" + Environment.NewLine + @"'). % ""\\\n""",
                Enumerable.Empty<string>(),
                new NTerm("writeq") {""}, // not quite sure why "" is to be expected, but it works
                expectConsumed: false);
        }

        [TestMethod]
        public void Test_8()
        {
            TestSuccess(
                @"writeq('\" + Environment.NewLine + @"a'). % ""\\\na""",
                Enumerable.Empty<string>(),
                new NTerm("writeq") {"a"}, // not quite sure why "a" is to be expected, but it works
                expectConsumed: false);
        }

        [TestMethod]
        public void Test_9()
        {
            TestSuccess(
                @"writeq('a\" + Environment.NewLine + @"b'). % ""a\\\nb""",
                Enumerable.Empty<string>(),
                new NTerm("writeq") {"ab"}, // not quite sure why "ab" is to be expected, but it works
                expectConsumed: false);
        }

        [TestMethod]
        public void Test_10()
        {
            TestSuccess(
                @"writeq('a\" + Environment.NewLine + @" b'). % ""a\\\n b""",
                Enumerable.Empty<string>(),
                new NTerm("writeq") {"a b"}, // not quite sure why "a b" is to be expected, but it works
                expectConsumed: false);
        }

        [TestMethod]
        public void Test_11()
        {
            TestError(@"writeq('\ ').", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_12()
        {
            TestError(@"writeq('\ '). % ""\\\t""", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_13()
        {
            TestSuccess(
                @"writeq('\t').",
                new NTerm("writeq") { "\x9" });
        }

        [TestMethod]
        public void Test_14()
        {
            TestSuccess(
                @"writeq('\a').",
                new NTerm("writeq") { "\x7" });
        }

        [TestMethod]
        public void Test_15()
        {
            TestSuccess(
                @"writeq('\7\').",
                new NTerm("writeq") { "\x7" });
        }

        [TestMethod]
        public void Test_16()
        {
            TestError(@"writeq('\ca').", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_17()
        {
            TestError(@"writeq('\e').", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_18()
        {
            TestSuccess(
                @"writeq('\033\').", 
                new NTerm("writeq") {"\x1B"});
        }

        [TestMethod]
        public void Test_19()
        {
            TestError(@"char_code('\e',C).", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_21()
        {
            TestError(@"char_code('\d',C).", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_22()
        {
            TestError(@"writeq('\u1').", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_23()
        {
            TestError(@"X = 0'\u1.", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_24()
        {
            TestError(@"writeq('", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_25()
        {
            TestError(@"writeq(.", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_26()
        {
            TestError(@"'\" + Environment.NewLine + @"''.", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_27()
        {
            TestSuccess(
                @"writeq((*)=(*)).",
                new NTerm("writeq") { new NTerm("=") { "*", "*" } });
        }

        [TestMethod]
        public void Test_28()
        {
            TestSuccess(
                @"writeq([:-,-]).", 
                new NTerm("writeq") { new NList {":-", "-"}});
        }

        // our own - not in the source document
        [TestMethod]
        public void Test_28a()
        {
            TestSuccess(@"writeq([(:-),-]).", new NTerm("writeq") { new NList { ":-", "-" } });
        }

        // our own - not in the source document
        [TestMethod]
        public void Test_28b()
        {
            TestSuccess(
                @"writeq([(:-),(-)]).",
                new NTerm("writeq") { new NList { ":-", "-" } });
        }

        [TestMethod]
        public void Test_29()
        {
            TestSuccess(
                @"writeq(f(*)).",
                new NTerm("writeq") { new NTerm("f") { "*" } });
        }

        [TestMethod]
        public void Test_30()
        {
            TestSuccess(
                @"writeq(a*(b+c)).",
                new NTerm("writeq") { new NTerm("*") { "a", new NTerm("+"){"b","c"} } });
        }

        [TestMethod]
        public void Test_31()
        {
            TestSuccess(
                @"writeq(f(;,'|',';;')).",
                new NTerm("writeq") { new NTerm("f") { ";", "|", ";;" } });
        }

        [TestMethod]
        public void Test_32()
        {
            TestSuccess(
                @"writeq([.,.(.,.,.)]).",
                new NTerm("writeq") { new NList { ".", new NTerm("."){".",".","."}}});
        }

        [TestMethod]
        public void Test_33()
        {
            TestSuccess(
                @"writeq((a :- b,c)).",
                new NTerm("writeq") { new NTerm(":-") { "a", new NTerm(",") { "b", "c" } } });
        }

        [TestMethod]
        public void Test_34()
        {
            TestSuccess(
                @"write_canonical([a]).",
                new NTerm("write_canonical") { new NList{"a"} });
        }

        [TestMethod]
        public void Test_35()
        {
            TestSuccess(
                @"writeq('/*').",
                new NTerm("writeq"){"/*"});
        }

        [TestMethod]
        public void Test_36()
        {
            TestSuccess(
                @"writeq('/**').",
                new NTerm("writeq") { "/**" });
        }

        [TestMethod]
        public void Test_37()
        {
            TestSuccess(
                @"writeq('*/').",
                new NTerm("writeq") { "*/" });
        }

        [TestMethod]
        public void Test_38()
        {
            TestSuccess(
                @"""\'\`\"""" = ""'`"""""". % """,
                new String[0],
                new NTerm("=") {new NCodes(@"'`"""), new NCodes(@"'`""")},
                expectConsumed:false);
        }

        [TestMethod]
        public void Test_39()
        {
            TestSuccess(
                @"'\'\`\""' = '''`""'.",
                new NTerm("=") { @"'`""", @"'`""" });
        }

        [TestMethod]
        public void Test_40()
        {
            TestSuccess(
                @"writeq('\'\`\""\""').",
                new NTerm("writeq") { @"'`""""" });
        }

        [TestMethod]
        public void Test_41()
        {
            TestSuccess(
                @"('\\') = (\).",
                new NTerm("=") { @"\", @"\" });
        }

        [TestMethod]
        public void Test_43()
        {
            TestError(@"X = 0X1.", TermSyntaxErrorCode.SpecificTokenExpected);
        }

        [TestMethod]
        public void Test_44()
        {
            TestError(@"float(.0).", TermSyntaxErrorCode.SpecificTokenExpected);
        }

        [TestMethod]
        public void Test_46()
        {
            TestError(@"float(- .0).", TermSyntaxErrorCode.SpecificTokenExpected);
        }

        [TestMethod]
        public void Test_47()
        {
            TestError(@"float(1E9).", TermSyntaxErrorCode.SpecificTokenExpected);
        }

        [TestMethod]
        public void Test_48()
        {
            TestError(@"integer(1e).", TermSyntaxErrorCode.SpecificTokenExpected);
        }

        [TestMethod]
        public void Test_53()
        {
            TestSuccess(
                @"writeq(1.0e100).",
                new NTerm("writeq") { 1.0e100 });
        }

        [TestMethod]
        public void Test_54()
        {
            TestError(@"float(1.0ee9).", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_56()
        {
            TestSuccess(
                @"integer(- 1).",
                new NTerm("integer") { -1 });
        }

        [TestMethod]
        public void Test_57()
        {
            TestSuccess(
                @"integer('-'1).",
                new NTerm("integer") { -1 });
        }

        [TestMethod]
        public void Test_58()
        {
            TestSuccess(
                @"integer('-' 1).",
                new NTerm("integer") { -1 });
        }

        [TestMethod]
        public void Test_59()
        {
            TestSuccess(
                @"integer(- /**/1).",
                new NTerm("integer") { -1 });
        }

        [TestMethod]
        public void Test_60()
        {
            TestError(@"integer(-/**/1).", TermSyntaxErrorCode.SpecificTokenExpected);
        }

        [TestMethod]
        public void Test_61()
        {
            TestSuccess(
                @"integer('-'/**/1).",
                new NTerm("integer") { -1 });
        }

        [TestMethod]
        public void Test_62()
        {
            TestSuccess(
                @"atom(-/**/-).",
                new NTerm("atom") {"-/**/-"});  // TODO: verify with standard - is it the right atom name here?
        }

        [TestMethod]
        public void Test_65()
        {
            TestSuccess(
                @"compound(+1).",
                new NTerm("compound") { new NTerm("+") { 1 } });
        }

        [TestMethod]
        public void Test_66()
        {
            TestSuccess(
                @"compound(+ 1).",
                new NTerm("compound") { new NTerm("+") {1}});
        }

        [TestMethod]
        public void Test_68()
        {
            TestSuccess(
                @"[(:-)|(:-)]=[:-|:-].",
                new NTerm("=")
                {
                    new NTerm(".") {":-", ":-"},
                    new NTerm(".") {":-", ":-"}
                });
        }

        [TestMethod]
        public void Test_69()
        {
            TestError(@"X=[a|b,c].", TermSyntaxErrorCode.SpecificTokenExpected);
        }

        [TestMethod]
        public void Test_75()
        {
            TestError(@"var(a:-b).", TermSyntaxErrorCode.SpecificTokenExpected);
        }

        [TestMethod]
        public void Test_76()
        {
            TestError(@":- = :- .", TermSyntaxErrorCode.UnexpectedInfixOperator);
        }

        [TestMethod]
        public void Test_77()
        {
            TestError(@"- = - .", TermSyntaxErrorCode.UnexpectedInfixOperator);
        }

        [TestMethod]
        public void Test_78()
        {
            TestError(@"* = * .", TermSyntaxErrorCode.UnexpectedInfixOperator);
        }

        [TestMethod]
        public void Test_81()
        {
            TestSuccess(
                @"{- - c}={-(-(c))}.",
                new NTerm("=")
                {
                    new NTerm("{}") { new NTerm("-") { new NTerm("-") {"c"}}},
                    new NTerm("{}") { new NTerm("-") { new NTerm("-") {"c"}}}
                });
        }

        [TestMethod]
        public void Test_82()
        {
            TestError(@"(- -) = -(-).", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_83()
        {
            TestError(@"(- - -) = -(-(-)).", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_84()
        {
            TestError(@"(- - - -) = -(-(-(-))).", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_85()
        {
            TestError(@"{:- :- c} = {:-(:-,c)}.", TermSyntaxErrorCode.MisplacedPrefixOperator);
        }

        [TestMethod]
        public void Test_86()
        {
            TestError(@"{- = - 1}={(-(=)) - 1}.", TermSyntaxErrorCode.UnexpectedInfixOperator);
        }

        // this appears identical to 88?
        [TestMethod]
        public void Test_87()
        {
            TestError(@"write_canonical((- = - 1)).", TermSyntaxErrorCode.UnexpectedInfixOperator);
        }

        [TestMethod]
        public void Test_88()
        {
            TestError(@"write_canonical((- = -1)).", TermSyntaxErrorCode.UnexpectedInfixOperator);
        }

        [TestMethod]
        public void Test_89()
        {
            TestError(@"write_canonical((-;)).", TermSyntaxErrorCode.UnexpectedInfixOperator);
        }

        [TestMethod]
        public void Test_90()
        {
            TestError(@"write_canonical((-;-)).", TermSyntaxErrorCode.UnexpectedInfixOperator);
        }

        [TestMethod]
        public void Test_91()
        {
            TestError(@"write_canonical((:-;-)).", TermSyntaxErrorCode.UnexpectedInfixOperator);
        }

        [TestMethod]
        public void Test_92()
        {
            TestError(@"[:- -c] = [(:- -c)].", TermSyntaxErrorCode.MisplacedPrefixOperator);
        }

        [TestMethod]
        public void Test_93()
        {
            TestError(@"writeq([a,b|,]).", TermSyntaxErrorCode.UnexpectedInfixOperator);
        }

        [TestMethod]
        public void Test_94()
        {
            TestError(@"X ={,}.", TermSyntaxErrorCode.UnexpectedInfixOperator);
        }

        [TestMethod]
        public void Test_95()
        {
            TestSuccess(
                @"{1} = {}(1).",
                new NTerm("=") { new NTerm("{}"){1}, new NTerm("{}"){1} });
        }

        [TestMethod]
        public void Test_96()
        {
            TestSuccess(
                @"write_canonical({1}).",
                new NTerm("write_canonical") { new NTerm("{}") {1}});
        }

        [TestMethod]
        public void Test_97()
        {
            var X = new Variable();
            TestSuccess(
                @"'[]'(1) = [ ](X).",
                new[] { "X" },
                new NTerm("=") { new NTerm("[]") { 1 }, new NTerm("[]") { X } });
        }

        // our own - not in the source document
        [TestMethod]
        public void Test_97a()
        {
            var X = new Variable();
            TestSuccess(
                @"'[]'(1) = [](X).",
                new[] { "X" },
                new NTerm("=") { new NTerm("[]") { 1 }, new NTerm("[]") { X } });
        }

        [TestMethod]
        public void Test_98()
        {
            TestError(@"X = [] (1).", TermSyntaxErrorCode.SpecificTokenExpected);
        }

        [TestMethod]
        public void Test_100()
        {
            TestSuccess(
                @"'''' = '\''.",
                new NTerm("=") { "'", "'" });
        }

        [TestMethod]
        public void Test_101()
        {
            TestSuccess(
                @"a = '\141\'.",
                new NTerm("=") { "a", "a" });
        }

        [TestMethod]
        public void Test_102()
        {
            TestError(@"a = '\141'.", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_103()
        {
            var X = new Variable();
            TestSuccess(
                @"X = '\141\141'.",
                new[] { "X" },
                new NTerm("=") { X, "a141" });
        }

        [TestMethod]
        public void Test_104()
        {
            TestError(@"X = '\9'.", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_105()
        {
            TestError(@"X = '\N'.", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_106()
        {
            TestError(@"X = '\\' .", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_107()
        {
            TestError(@"X = '\77777777777\'.", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_108()
        {
            TestSuccess(
                @"a = '\x61\'.",
                new NTerm("=") { "a", "a" });
        }

        [TestMethod]
        public void Test_109()
        {
            TestError(@"atom_codes('\xG\',Cs).", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_110()
        {
            TestError(@"atom_codes('\xG1\',Cs).", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_111()
        {
            TestError(@"atom(`).", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_112()
        {
            TestError(@"atom(`+).", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_113()
        {
            var X = new Variable();
            TestSuccess(
                @"X = `a`.",
                new[] { "X" },
                new NTerm("=") { X, new PString("a") });
        }

        [TestMethod]
        public void Test_114()
        {
            TestSuccess(
                @"integer(0'\').",
                new NTerm("integer") { 39 });
        }

        [TestMethod]
        public void Test_115()
        {
            TestSuccess(
                @"integer(0''').",
                new NTerm("integer") { 39 });
        }

        [TestMethod]
        public void Test_116()
        {
            TestSuccess(
                @"0''' = 0'\'.",
                new NTerm("=") { 39, 39 });
        }

        [TestMethod]
        public void Test_117()
        {
            TestError(@"integer(0'').", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_121()
        {
            TestError(@"X = 2'1.", TermSyntaxErrorCode.SpecificTokenExpected);
        }

        [TestMethod]
        public void Test_123()
        {
            var X = new Variable();
            TestSuccess(
                @"X =0'\x41\ .",
                new [] {"X"},
                new NTerm("=") {X, 65});
        }

        [TestMethod]
        public void Test_124()
        {
            var X = new Variable();
            TestSuccess(
                @"X =0'\x41\.",
                new[] { "X" },
                new NTerm("=") { X, 65 });
        }

        [TestMethod]
        public void Test_125()
        {
            var X = new Variable();
            TestSuccess(
                @"X =0'\x1\.",
                new[] { "X" },
                new NTerm("=") { X, 1 });
        }

        [TestMethod]
        public void Test_126()
        {
            TestWait(
                @"X = 0'\" + Environment.NewLine + @".\");
        }

        [TestMethod]
        public void Test_127()
        {
            var X = new Variable();
            TestSuccess(
                @"X is 16'mod'2.",
                new[] { "X" },
                new NTerm("is") { X, new NTerm("mod") {16,2}});
        }

        [TestMethod]
        public void Test_128()
        {
            var X = new Variable();
            TestSuccess(
                @"X is 37'mod'2.",
                new[] { "X" },
                new NTerm("is") { X, new NTerm("mod") { 37, 2 } });
        }

        [TestMethod]
        public void Test_129()
        {
            TestError(@"X is 0'mod'1.", TermSyntaxErrorCode.SpecificTokenExpected);
        }

        [TestMethod]
        public void Test_130()
        {
            var X = new Variable();
            TestSuccess(
                @"X is 1'+'1.",
                new[] { "X" },
                new NTerm("is") { X, new NTerm("+") { 1, 1 } });
        }

        [TestMethod]
        public void Test_131()
        {
            TestSuccess(
                @"op(100,fx,' op').",
                new NTerm("op") { 100, "fx", " op"});
        }

        [TestMethod]
        public void Test_135()
        {
            TestSuccess(
                @"writeq(-(1)).",
                new NTerm("writeq") { new NTerm("-") { 1 } });
        }

        [TestMethod]
        public void Test_137()
        {
            TestSuccess(
                @"writeq(- (a*b)).",
                new NTerm("writeq") { new NTerm("-") { new NTerm("*") {"a", "b"}}});
        }

        [TestMethod]
        public void Test_138()
        {
            TestSuccess(
                @"writeq(\ (a*b)).",
                new NTerm("writeq") { new NTerm(@"\") { new NTerm("*") { "a", "b" } } });
        }

        [TestMethod]
        public void Test_139()
        {
            TestSuccess(
                @"writeq(-((a,b))).",
                new NTerm("writeq") { new NTerm("-") { new NTerm(",") { "a", "b" } } });
        }

        [TestMethod]
        public void Test_140()
        {
            TestSuccess(
                @"writeq(-a).",
                new NTerm("writeq") { new NTerm("-") { "a" } });
        }

        [TestMethod]
        public void Test_141()
        {
            var P = new Variable();
            TestSuccess(
                @"current_op(P,xfy,.).",
                new[] { "P" },
                new NTerm("current_op") { P, "xfy", "." });
        }

        [TestMethod]
        public void Test_142()
        {
            TestSuccess(
                @"op(400,xfy,.).",
                new NTerm("op") { 400, "xfy", "." });
        }

        [TestMethod]
        public void Test_165()
        {
            TestSuccess(
                @"atom('.''-''.').",
                new NTerm("atom") { ".'-'." });
        }

        [TestMethod]
        public void Test_172()
        {
            var X = new Variable();
            TestSuccess(
                @"X is 10.0** -323, writeq(X).",
                new[] { "X" },
                new NTerm(",") { new NTerm("is") {X, new NTerm("**") { 10.0, -323 } }, new NTerm("writeq") {X} });
        }

        [TestMethod]
        public void Test_173()
        {
            TestSuccess(
                @"1.0e-323=:=10.0** -323.",
                new NTerm("=:=") { 1.0e-323, new NTerm("**") {10.0, -323}});
        }

        [TestMethod]
        public void Test_174()
        {
            TestSuccess(
                @"-1 = -0x1.",
                new NTerm("=") { -1, -1 });
        }

        [TestMethod]
        public void Test_175()
        {
            var T = new Variable();
            TestSuccess(
                @"T = t(0b1,0o1,0x1).",
                new[] { "T" },
                new NTerm("=") { T, new NTerm("t") { 1, 1, 1 } });
        }

        [TestMethod]
        public void Test_176()
        {
            var X = new Variable();
            TestSuccess(
                @"X is 0b1mod 2.",
                new[] { "X" },
                new NTerm("is") { X, new NTerm("mod") {1, 2}});
        }

        [TestMethod]
        public void Test_177()
        {
            TestError(@"0'\t=0'" + "\x9" + @" . % horiz. tab ", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_178()
        {
            TestSuccess(
                @"""\`"" = ""`"".",
                new NTerm("=") { new NCodes(@"`"), new NCodes(@"`") });
        }

        [TestMethod]
        public void Test_179()
        {
            TestSuccess(
                @"""\'\"""" = ""'"""""". % """,
                new String[0],
                new NTerm("=") {new NCodes(@"'"""), new NCodes(@"'""")},
                expectConsumed:false);
        }

        [TestMethod]
        public void Test_182()
        {
            TestSuccess(
                @"writeq(-(-1)).",
                new NTerm("writeq") { new NTerm("-") { -1 } });
        }

        [TestMethod]
        public void Test_183()
        {
            TestSuccess(
                @"writeq(-(1^2)).",
                new NTerm("writeq") { new NTerm("-") { new NTerm("^") { 1, 2 } } });
        }

        [TestMethod]
        public void Test_184()
        {
            TestSuccess(
                @"writeq(-(-)).",
                new NTerm("writeq") { new NTerm("-") { "-" } });
        }

        [TestMethod]
        public void Test_185()
        {
            TestSuccess(
                @"writeq(-[-]).",
                new NTerm("writeq") { new NTerm("-") { new NList {"-"} }});
        }

        [TestMethod]
        public void Test_186()
        {
            var X = new Variable();
            TestSuccess(
                @"X/* /*/=7.",
                new[] { "X" },
                new NTerm("=") { X, 7 });
        }

        [TestMethod]
        public void Test_187()
        {
            var X = new Variable();
            TestSuccess(
                @"X/*/*/=7.",
                new[] { "X" },
                new NTerm("=") { X, 7 });
        }

        [TestMethod]
        public void Test_188()
        {
            TestSuccess(
                @"writeq(-p(c)).",
                new NTerm("writeq") { new NTerm("-") { new NTerm("p") { "c" } } });
        }

        [TestMethod]
        public void Test_189()
        {
            TestSuccess(
                @"writeq(-{}).",
                new NTerm("writeq") { new NTerm("-") { new NTerm("{}") } });
        }

        [TestMethod]
        public void Test_190()
        {
            TestSuccess(
                @"writeq(-{a}).",
                new NTerm("writeq") { new NTerm("-") { new NTerm("{}") { "a" } } });
        }

        [TestMethod]
        public void Test_191()
        {
            TestSuccess(
                @"writeq(-(-a)).",
                new NTerm("writeq") { new NTerm("-") { new NTerm("-") { "a" } } });
        }

        [TestMethod]
        public void Test_192()
        {
            TestSuccess(
                @"writeq(-(-(-a))).",
                new NTerm("writeq") { new NTerm("-") { new NTerm("-") { new NTerm("-") {"a"}} } });
        }

        [TestMethod]
        public void Test_193()
        {
            TestError(@"writeq('\ " + Environment.NewLine + @"'). % ""\\ \n"" ", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_194()
        {
            TestSuccess(
                @"writeq((.)+(.)).",
                new NTerm("writeq") {new NTerm("+") { ".", "." }});
        }

        [TestMethod]
        public void Test_195()
        {
            TestSuccess(
                @"op(100,xf,'').",
                new NTerm("op") { 100, "xf", "" });
        }

        [TestMethod]
        public void Test_198()
        {
            TestSuccess(
                @"atom($-).",
                new NTerm("atom") { "$-" });
        }

        [TestMethod]
        public void Test_199()
        {
            TestSuccess(
                @"atom(-$).",
                new NTerm("atom") { "-$" });
        }

        [TestMethod]
        public void Test_203()
        {
            TestSuccess(
                @"writeq(//*).",
                new NTerm("writeq") { "//*" });
        }

        [TestMethod]
        public void Test_210()
        {
            TestError(@"X = 0'\.", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_211()
        {
            TestError(@"X = 0'\. .", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_212()
        {
            var X = new Variable();
            TestSuccess(
                @"X is 1'\" + Environment.NewLine + @"+'1.",
                new[] { "X" },
                new NTerm("is") { X, new NTerm("+") { 1, 1 } });
        }

        [TestMethod]
        public void Test_213()
        {
            var X = new Variable();
            TestSuccess(
                @"X is 0'\" + Environment.NewLine + @"+'1.",
                new[] { "X" },
                new NTerm("is") { X, new NTerm("+") { 0, 1 } });
        }

        [TestMethod]
        public void Test_214()
        {
            TestWait(@"X is 0'\");
        }

        [TestMethod]
        public void Test_216()
        {
            TestSuccess(
                @"writeq(-(-(1))).",
                new NTerm("writeq") { new NTerm("-") { new NTerm("-") {1} }});
        }

        [TestMethod]
        public void Test_218()
        {
            TestSuccess(
                @"writeq(-(1*2)).",
                new NTerm("writeq") { new NTerm("-") { new NTerm("*") {1,2}} });
        }

        [TestMethod]
        public void Test_222()
        {
            TestSuccess(
                @"writeq((-)-(-)).",
                new NTerm("writeq") { new NTerm("-") {"-","-"}});
        }

        [TestMethod]
        public void Test_223()
        {
            TestSuccess(
                @"writeq(((:-):-(:-))).",
                new NTerm("writeq") { new NTerm(":-") { ":-", ":-" } });
        }

        [TestMethod]
        public void Test_224()
        {
            TestError(@"\ .", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_225()
        {
            var C = new Variable();
            TestSuccess(
                @"char_code(C,0), writeq(C).",
                new[] { "C" },
                new NTerm(",") { new NTerm("char_code") { C, 0 }, new NTerm("writeq") {C}});
        }

        [TestMethod]
        public void Test_226()
        {
            var _1 = new Variable();
            var _2 = new Variable();
            TestSuccess(
                @"write_canonical(_+_).",
                new NTerm("write_canonical") { new NTerm("+") { _1, _2 } });
        }

        [TestMethod]
        public void Test_227()
        {
            var B = new Variable();
            TestSuccess(
                @"write_canonical(B+B).",
                new[] { "B" },
                new NTerm("write_canonical") { new NTerm("+") { B, B } });
        }

        [TestMethod]
        public void Test_228()
        {
            TestError(@"writeq(0'\z).", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_230()
        {
            TestError(@"char_code('\^',X).", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_231()
        {
            TestError(@"writeq(0'\c).", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_232()
        {
            TestError(@"writeq(0'\ ).", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_233()
        {
            TestError(@"writeq(nop (1)).", TermSyntaxErrorCode.SpecificTokenExpected);
        }

        [TestMethod]
        public void Test_236()
        {
            TestSuccess(
                @"write_canonical(a- - -b).",
                new NTerm("write_canonical") { new NTerm("-") {"a", new NTerm("-") {new NTerm("-") {"b"}}}});
        }

        [TestMethod]
        public void Test_241()
        {
            TestError(@"writeq('\d').", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_250()
        {
            TestSuccess(
                @"writeq('\0\').",
                new NTerm("writeq") { "\0" });
        }

        [TestMethod]
        public void Test_257()
        {
            TestSuccess(
                @"writeq([+{a},+[]]).",
                new NTerm("writeq") { new NList
                {
                    new NTerm("+") { new NTerm("{}") {"a"} },
                    new NTerm("+") { "[]"} 
                }});
        }

        [TestMethod]
        public void Test_259()
        {
            var X = new Variable();
            TestSuccess(
                @"X = 0'\" + Environment.NewLine + @"+'/*'. %*/1.",
                new[] { "X" },
                new NTerm("=") { X, new NTerm("+") { 0, 1 } });
        }

        [TestMethod]
        public void Test_260()
        {
            TestSuccess(
                @"writeq(-(a^2)).",
                new NTerm("writeq") { new NTerm("-") { new NTerm("^") { "a", 2 } } });
        }

        [TestMethod]
        public void Test_261()
        {
            TestError(@")" + Environment.NewLine + @"'", TermSyntaxErrorCode.InvalidSegmentStart);
        }

        [TestMethod]
        public void Test_264()
        {
            TestSuccess(
                @"writeq('^`').",
                new NTerm("writeq") { "^`" });
        }

        [TestMethod]
        public void Test_269()
        {
            TestSuccess(
                @"writeq('\a\b\r\f\t\n').",
                new NTerm("writeq") { "\x07\x08\x0D\x0C\x09\x0A" });
        }

        [TestMethod]
        public void Test_270()
        {
            TestError(@"Finis ().", TermSyntaxErrorCode.SpecificTokenExpected);
        }

        // IMPROVE: have more - all of them

        // ReSharper restore InconsistentNaming
        #endregion
    }
}
