﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Linq;
using System.Windows.Forms;
using JetBrains.Annotations;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.OperationModel;

namespace OpcLabs.EasyOpcFormsTestApp.UA
{
    public partial class ChallengeTestForm : Form
    {
        private class TestParameters
        {
            public bool EnableOpcFoundationLocal = false;
            public bool EnableSoftingLocal = false;

            public bool EnableBeckhoffTf6100 = true;
            public bool EnableBoschIndraControl = true;
            public bool EnableProgeaMoviconNext = true;
            public bool EnableProsysJava2 = true;
            public bool EnableSiemensS7 = true;
            public bool EnableSiemensS7Opt = true;
            public bool EnableSoftingNetSdk = true;
            public bool EnableSoftingToolbox = true;
        }

        public ChallengeTestForm()
        {
            InitializeComponent();

            //panel1.DoubleBuffered = true;

            RecalculateGeometry();
        }

        public int ElementCount
        {
            get { return _elementCount; }
            private set 
            { 
                int oldValue = _elementCount;
                _elementCount = value;
                if (value != oldValue)
                    OnElementCountChanged(oldValue);
            }
        }

        private Point DimensionElements { get; set; }
        private PointF ElementSize { get; set; }

        static private Color ColorFromHashCode(int hashCode)
        {
            hashCode = unchecked((int)(hashCode*2654435761));
            int argb = (hashCode | unchecked((int) 0xFF000000)) | (hashCode >> 24);
            return Color.FromArgb(argb);
        }

        private Rectangle ComputeElementRectangle(int elementIndex)
        {
            // ReSharper disable SuggestUseVarKeywordEvident
            int positionY = elementIndex/DimensionElements.X;
            int positionX = elementIndex - DimensionElements.X*positionY;
            int x = (int) (positionX * ElementSize.X);
            int y = (int)(positionY * ElementSize.Y);
            int nextX = (int)((positionX + 1) * ElementSize.X);
            int nextY = (int)((positionY + 1) * ElementSize.Y);

            Rectangle rectangle = new Rectangle(x, y, nextX - x, nextY - y);
            return rectangle;
            // ReSharper restore SuggestUseVarKeywordEvident
        }

        [NotNull]
        private IEnumerable<EasyUAMonitoredItemArguments> CreateMonitoredItems([NotNull] TestParameters testParameters)
        {
            const float loadFactor = 1.0f/1;
            const int samplingInterval = 250;

            UAMonitoringParameters monitoringParameters = samplingInterval;

            var list = new List<EasyUAMonitoredItemArguments>();

            // Beckhoff TF6100 OPC-UA 2.0.0.8
            if (testParameters.EnableBeckhoffTf6100)
            {
                UAEndpointDescriptor endpointDescriptor = "opc.tcp://192.168.10.168:4841";
                string[] names =
                    {
                        "dyn_BOOL", "dyn_BOOLx",
                        "dyn_BSTR_0", "dyn_BSTR_1", "dyn_BSTR_2", "dyn_BSTRARRAY",
                        "dyn_I1", "dyn_I1x", "dyn_I2", "dyn_I2x", "dyn_I4", "dyn_I4x",
                        "dyn_R4", "dyn_R4x", "dyn_R8", "dyn_R8x",
                        "dyn_UI1", "dyn_UI1x", "dyn_UI2", "dyn_UI2x", "dyn_UI4", "dyn_UI4x"
                    };
                foreach (string name in names)
                    list.Add(new EasyUAMonitoredItemArguments(null, endpointDescriptor,
                        String.Format("nsu=PLC1;ns=4;s=GVL_dynamic.{0}", name),
                        monitoringParameters));
            }

            // Bosch Rexroth AG - IndraControl MLC L65 13VRS
            if (testParameters.EnableBoschIndraControl)
            {
                UAEndpointDescriptor endpointDescriptor = "opc.tcp://192.168.10.35";
                string[] names =
                    {
                        "bBool", "byByte", "diDint", "dwDate", "dwDateTime", "dwDword", "dwTime", "dwTimeDay", "dwTod",
                        "iInt", "liLint", "lrLreal", "lwLtime", "lwLword", "rReal", "siSint", "strString", "udiUdint",
                        "uiUint", "uliUlint", "usiUsint", "wstrWstring", "wWord"
                    };
                foreach (string name in names)
                    list.Add(new EasyUAMonitoredItemArguments(null, endpointDescriptor,
                        String.Format("nsu=http://www.boschrexroth.com/IndraLogic;ns=4;s=Application.DYNDATA.{0}", name),
                        monitoringParameters));
            }

            // OPC Foundation (local)
            if (testParameters.EnableOpcFoundationLocal)
            {
                UAEndpointDescriptor endpointDescriptor = "opc.tcp://localhost:51210/UA/SampleServer";
                for (int i = 0; i < loadFactor * 100; i++)
                    list.Add(new EasyUAMonitoredItemArguments(null, endpointDescriptor,
                        String.Format("nsu=http://samples.org/UA/memorybuffer/Instance;s=UInt32[{0}]", i),
                        monitoringParameters));
            }

            // Progea Movicon.NExT 12.0.8
            if (testParameters.EnableProgeaMoviconNext)
            {
                UAEndpointDescriptor endpointDescriptor = "opc.tcp://notebook10:62000/MovNextUAServer";
                for (int i = 1; i <= loadFactor * 100; i++)
                    list.Add(new EasyUAMonitoredItemArguments(null, endpointDescriptor,
                        String.Format(
                            "ns=2;s=28f9ab99-04c9-45f1-9d9e-a5b1d95fa1dd?DynamicValues/Scalar/ReadWrite/Booleans/Boolean{0:D3}",
                            i),
                        monitoringParameters));
                for (int i = 1; i <= loadFactor * 100; i++)
                    list.Add(new EasyUAMonitoredItemArguments(null, endpointDescriptor,
                        String.Format(
                            "ns=2;s=28f9ab99-04c9-45f1-9d9e-a5b1d95fa1dd?DynamicValues/Scalar/ReadWrite/Int32/Int32_{0:D3}",
                            i),
                        monitoringParameters));
                for (int i = 1; i <= loadFactor * 100; i++)
                    list.Add(new EasyUAMonitoredItemArguments(null, endpointDescriptor,
                        String.Format(
                            "ns=2;s=28f9ab99-04c9-45f1-9d9e-a5b1d95fa1dd?DynamicValues/Scalar/ReadWrite/UInt64/UInt64_{0:D3}",
                            i),
                        monitoringParameters));
                for (int i = 1; i <= loadFactor * 100; i++)
                    list.Add(new EasyUAMonitoredItemArguments(null, endpointDescriptor,
                        String.Format(
                            "ns=2;s=28f9ab99-04c9-45f1-9d9e-a5b1d95fa1dd?DynamicValues/Scalar/ReadWrite/Double/Double{0:D3}",
                            i),
                        monitoringParameters));
                for (int i = 1; i <= loadFactor * 100; i++)
                    list.Add(new EasyUAMonitoredItemArguments(null, endpointDescriptor,
                        String.Format(
                            "ns=2;s=28f9ab99-04c9-45f1-9d9e-a5b1d95fa1dd?DynamicValues/Scalar/ReadWrite/String/String{0:D3}",
                            i),
                        monitoringParameters));
                // ...
            }

            // Prosys OPC UA Java Server 2.0 BETA
            if (testParameters.EnableProsysJava2)
            {
                UAEndpointDescriptor endpointDescriptor = "opc.tcp://prosys01:53530/OPCUA/SimulationServer";
                for (int i = 0; i < loadFactor * 1000; i++)
                    list.Add(new EasyUAMonitoredItemArguments(null, endpointDescriptor,
                        String.Format("ns=4;s=DataItem_{0:D4}", i),
                        monitoringParameters));
            }

            // Siemens SIMATIC NET OPC Server S7 V12
            if (testParameters.EnableSiemensS7)
            {
                UAEndpointDescriptor endpointDescriptor = "opc.tcp://SiemensOPC-S7:4845";
                for (int i = 0; i < loadFactor*100; i++)
                {
                    list.Add(new EasyUAMonitoredItemArguments(null, endpointDescriptor,
                        String.Format("nsu=SYM:;ns=7;s=S7_400.Dynamic.Scalar.ReadWrite.Bool{0:D3}", i + 1),
                        monitoringParameters));
                    list.Add(new EasyUAMonitoredItemArguments(null, endpointDescriptor,
                        String.Format("nsu=SYM:;ns=7;s=S7_400.Dynamic.Scalar.ReadWrite.Float{0:D3}", i + 1),
                        monitoringParameters));
                    list.Add(new EasyUAMonitoredItemArguments(null, endpointDescriptor,
                        String.Format("nsu=SYM:;ns=7;s=S7_400.Dynamic.Scalar.ReadWrite.Int16{0:D3}", i + 1),
                        monitoringParameters));
                    list.Add(new EasyUAMonitoredItemArguments(null, endpointDescriptor,
                        String.Format("nsu=SYM:;ns=7;s=S7_400.Dynamic.Scalar.ReadWrite.String{0:D3}", i + 1),
                        monitoringParameters));
                }
            }

            // Siemens SIMATIC NET OPC Server S7 opt V12
            if (testParameters.EnableSiemensS7Opt)
            {
                UAEndpointDescriptor endpointDescriptor = "opc.tcp://SiemensOPC-S7:4850";
                for (int i = 0; i < loadFactor * 100; i++)
                {
                    list.Add(new EasyUAMonitoredItemArguments(null, endpointDescriptor,
                        String.Format("nsu=SYM:;ns=7;s=S7_1500.Dynamic.Scalar.ReadWrite.Bool{0:D3}", i + 1),
                        monitoringParameters));
                    list.Add(new EasyUAMonitoredItemArguments(null, endpointDescriptor,
                        String.Format("nsu=SYM:;ns=7;s=S7_1500.Dynamic.Scalar.ReadWrite.Float{0:D3}", i + 1),
                        monitoringParameters));
                    list.Add(new EasyUAMonitoredItemArguments(null, endpointDescriptor,
                        String.Format("nsu=SYM:;ns=7;s=S7_1500.Dynamic.Scalar.ReadWrite.Int16{0:D3}", i + 1),
                        monitoringParameters));
                    list.Add(new EasyUAMonitoredItemArguments(null, endpointDescriptor,
                        String.Format("nsu=SYM:;ns=7;s=S7_1500.Dynamic.Scalar.ReadWrite.String{0:D3}", i + 1),
                        monitoringParameters));
                }
            }

            // Softing UA Demo Server (local)
            if (testParameters.EnableSoftingLocal)
            {
                UAEndpointDescriptor endpointDescriptor = "opc.tcp://localhost:51510/UA/DemoServer";
                for (int i = 0; i < loadFactor * 100; i++)
                    list.Add(new EasyUAMonitoredItemArguments(null, endpointDescriptor,
                        String.Format(
                            "nsu=http://opcfoundation.org/Quickstarts/ReferenceApplications;s=/Dynamic/All Profiles/Scalar Mass/Int32/Int32{0:D3}",
                            i),
                        monitoringParameters));
            }

            // Softing OPC UA .NET SDK Server (OPC UA Demo Server)
            if (testParameters.EnableSoftingNetSdk)
            {
                UAEndpointDescriptor endpointDescriptor = "opc.tcp://wdelegw7-pc:62541/TestServer";
                for (int i = 0; i < loadFactor * 1000; i++)
                    list.Add(new EasyUAMonitoredItemArguments(null, endpointDescriptor,
                        String.Format("ns=8;i={0}", i + 1),
                        monitoringParameters));
            }

            // Softing OPC Toolbox UA Server for Windows
            if (testParameters.EnableSoftingToolbox)
            {
                UAEndpointDescriptor endpointDescriptor = "opc.tcp://WMEWIN7:4880/Softing";
                for (int i = 4; i <= 13; i++)
                    for (int j = 0; j < loadFactor*100; j++)
                        list.Add(new EasyUAMonitoredItemArguments(null, endpointDescriptor,
                            String.Format("ns={0};i={1}", i, j + 1),
                            monitoringParameters));
            }

            return list;
        }

        // ReSharper disable InconsistentNaming
        private void easyUAClient1_DataChangeNotification(object sender, [NotNull] EasyUAMonitoredItemChangedEventArgs e)
        // ReSharper restore InconsistentNaming
        {
            EasyUAMonitoredItemArguments arguments = e.Arguments;
            Debug.Assert(arguments != null);
            object state = arguments.State;
            Debug.Assert(state != null);
            // ReSharper disable SuggestUseVarKeywordEvident
            int elementIndex = (int)state;
            // ReSharper restore SuggestUseVarKeywordEvident

            _elementArray[elementIndex] = e;

            Rectangle elementRectangle = ComputeElementRectangle(elementIndex);
            panel1.Invalidate(elementRectangle);
        }

        private void OnElementCountChanged(
            // ReSharper disable UnusedParameter.Local
            int oldValue)
            // ReSharper restore UnusedParameter.Local
        {
            RecalculateGeometry();
        }

        private void PaintElement(int elementIndex, [NotNull] Graphics graphics)
        {
            Rectangle elementRectangle = ComputeElementRectangle(elementIndex);
            PaintElement(elementIndex, graphics, elementRectangle);
        }

        private void PaintElement(int elementIndex, [NotNull] Graphics graphics, Rectangle elementRectangle)
        {
            object element = _elementArray[elementIndex];

            graphics.DrawRectangle(new Pen(Color.Black), 
                elementRectangle.X, elementRectangle.Y, elementRectangle.Width - 1, elementRectangle.Height - 1);
            Rectangle innerRectangle = elementRectangle;
            innerRectangle.Inflate(-1, -1);

            if (element == null)
            {
                graphics.FillRectangle(new HatchBrush(HatchStyle.Horizontal, Color.DarkGray), innerRectangle);
                return;
            }

            var eventArgs = (EasyUADataChangeNotificationEventArgs) element;
            Exception exception = eventArgs.Exception;
            if (exception != null)
            {
                int hashCode = exception.GetHashCode();
                graphics.FillRectangle(new HatchBrush(HatchStyle.Vertical, ColorFromHashCode(hashCode)), innerRectangle);
                return;
            }

            UAAttributeData attributeData = eventArgs.AttributeData;
            Debug.Assert(attributeData != null);
            object value = attributeData.Value;
            if (value == null)
            {
                var hashCode = unchecked((int)attributeData.StatusCode.InternalValue);
                graphics.FillRectangle(new HatchBrush(HatchStyle.LightUpwardDiagonal, ColorFromHashCode(hashCode)), innerRectangle);
                return;
            }

            // We are intentionally ignoring the timestamps
            // IMPROVE: optionally, consider the whole UAAttributeData
            graphics.FillRectangle(new SolidBrush(ColorFromHashCode(value.GetHashCode())), innerRectangle);
        }

        private void PaintUnused(int elementIndex, [NotNull] Graphics graphics)
        {
            Rectangle elementRectangle = ComputeElementRectangle(elementIndex);
            Brush brush = new HatchBrush(HatchStyle.DiagonalCross, Color.DarkGray);
            graphics.FillRectangle(brush, elementRectangle);
        }

        // ReSharper disable InconsistentNaming
        private void panel1_Paint(object sender, [NotNull] PaintEventArgs e)
        // ReSharper restore InconsistentNaming
        {
            for (int i = 0; i < ElementCount; i++)
                PaintElement(i, e.Graphics);
            for (int i = ElementCount; i < DimensionElements.X*DimensionElements.Y; i++)
                PaintUnused(i, e.Graphics);
        }

        // ReSharper disable InconsistentNaming
        private void panel1_Resize(object sender, EventArgs e)
        // ReSharper restore InconsistentNaming
        {
            RecalculateGeometry();
            panel1.Invalidate();
        }

        private void RecalculateGeometry()
        {
            // ReSharper disable SuggestUseVarKeywordEvident
            int totalWidth = panel1.Width;
            int totalHeight = panel1.Height;

            int elementsHorizontally;
            int elementsVertically;
            if (ElementCount == 0)
            {
                elementsHorizontally = 1;
                elementsVertically = 1;
            }
            else
            {
                long totalArea = totalWidth * totalHeight;
                long elementArea = totalArea / ElementCount;

                float maximumElementWidth = (float)Math.Sqrt(elementArea);
                float maximumElementHeight = maximumElementWidth;

                elementsHorizontally = (int)Math.Ceiling((decimal)(totalWidth / maximumElementWidth));
                elementsVertically = (int)Math.Ceiling((decimal)(totalHeight / maximumElementHeight));
            }

            float elementWidth = (float)totalWidth / elementsHorizontally;
            float elementHeight = (float)totalHeight / elementsVertically;

            DimensionElements = new Point(elementsHorizontally, elementsVertically);
            ElementSize = new PointF(elementWidth, elementHeight);
            // ReSharper restore SuggestUseVarKeywordEvident
        }

        // ReSharper disable InconsistentNaming
        private void startButton_Click(object sender, EventArgs e)
        // ReSharper restore InconsistentNaming
        {
            //easyUAClient1.IsolatedParameters.SessionParameters.CertificateAcceptancePolicy.AcceptAnyCertificate = true;
            easyUAClient1.IsolatedParameters.SessionParameters.CheckEndpointDomain = false;

            var testParameters = new TestParameters();

            IEnumerable<EasyUAMonitoredItemArguments> monitoredItems = CreateMonitoredItems(testParameters);
            EasyUAMonitoredItemArguments[] argumentsArray = monitoredItems.ToArray();
            int count = argumentsArray.Length;
            for (int i = 0; i < count; i++)
            {
                EasyUAMonitoredItemArguments arguments = argumentsArray[i];
                Debug.Assert(arguments != null);
                arguments.State = i;
            }

            Array.Resize(ref _elementArray, count);
            ElementCount = count;

            easyUAClient1.SubscribeMultipleMonitoredItems(argumentsArray);

            startButton.Enabled = false;
            stopButton.Enabled = true;
        }

        // ReSharper disable InconsistentNaming
        private void stopButton_Click(object sender, EventArgs e)
        // ReSharper restore InconsistentNaming
        {
            easyUAClient1.UnsubscribeAllMonitoredItems();

            startButton.Enabled = true;
            stopButton.Enabled = false;
        }

        [NotNull]
        private object[] _elementArray = new object[0];

        private int _elementCount;
    }
}
